import 'dart:async';
import 'dart:typed_data';
import 'dart:ui';

import 'package:cached_network_marker/src/params/circle_avatar_params.dart';
import 'package:flutter_cache_manager/flutter_cache_manager.dart';

/// The generator of custom marker. To show the custom marker on [Google map for flutter](https://pub.dev/packages/google_maps_flutter), you have to pass the bitmap to [BitmapDescriptor](https://pub.dev/documentation/google_maps_flutter_platform_interface/latest/google_maps_flutter_platform_interface/BitmapDescriptor-class.html).
/// This class will generate the custom marker with specified image url, and returns the bitmap.
/// You can only use [circleAvatar] for now, but maybe more shapes will be added in the future.
/// All marker bitmaps will be cached on user local for a month.
///
/// Your image will be resized automatically with original aspect ratio. The size of image will be decided with params for each generators.
class CachedNetworkMarker {
  /// Creates the marker generator. Provide the device pixel ratio by `MediaQuery.of(context).devicePixelRatio`
  CachedNetworkMarker({
    required this.url,
    required this.dpr,
    DefaultCacheManager? cm,
  }) : cacheManager = cm ?? DefaultCacheManager();

  /// The url of target image.
  final String url;

  /// Device pixel ratio of target device.
  final double dpr;

  /// CacheManager instance
  final DefaultCacheManager cacheManager;

  /// Intermidiate image descriptor
  ImageDescriptor? _descriptor;

  /// Aspect ratio of downloaded image.
  double get _aspectRatio => _descriptor!.width / _descriptor!.height;

  /// Initializes the image descriptor with provided image url.
  Future<void> _initImageDescriptor() async {
    final file = await cacheManager.getSingleFile(url);
    final bytes = await file.readAsBytes();
    _descriptor = await ImageDescriptor.encoded(
      await ImmutableBuffer.fromUint8List(bytes),
    );
  }

  /// Get and resize the image from created image descriptor.
  Future<Image> _getImage(int diameter) async {
    final imageCodec = _aspectRatio >= 1
        ? await _descriptor!.instantiateCodec(targetHeight: diameter)
        : await _descriptor!.instantiateCodec(targetWidth: diameter);
    final frame = await imageCodec.getNextFrame();
    return frame.image;
  }

  /// Build the cache key from provided key prefix.
  String _buildKey(String prefix) {
    return '$prefix/$url';
  }

  /// Retrieve file from local cache. If file does not exists, it will return null.
  Future<Uint8List?> _getFileFromCache(String key) async {
    final cache = await cacheManager.getFileFromCache(key);
    if (cache != null) {
      return cache.file.readAsBytes();
    } else {
      return null;
    }
  }

  /// Generate the circle avatar marker bitmap.
  /// Specify the color and sizes of marker with [CircleAvatarParams].
  /// The image will be resized to diameter of circle([CircleAvatarParams.radius] * [dpr] * 2) with original aspect ratio.
  /// It will automatically clip the center of image so the overflowed part won't be showed.
  Future<Uint8List> circleAvatar(CircleAvatarParams params) async {
    final cacheKey = _buildKey(params.keyPrefix);
    final cache = await _getFileFromCache(cacheKey);
    if (cache != null) {
      return cache;
    }

    final radius = params.radius * dpr;
    final borderWidth = params.borderWidth * dpr;
    final dotRadius = params.dotRadius * dpr;
    final diameter = radius * 2;
    final dotDiameter = dotRadius * 2;
    final spacing = params.spacing * dpr;

    await _initImageDescriptor();
    final image = await _getImage(diameter.toInt());

    final imageStartingPoint = _aspectRatio >= 1
        ? Offset(0 - ((image.width - diameter) / 2), 0)
        : Offset(0, 0 - ((image.height - diameter) / 2));
    final imageCenter = Offset(radius, radius);

    final recorder = PictureRecorder();
    final canvas = Canvas(recorder);

    final outlinePath = Path()
      ..addOval(Rect.fromCircle(center: imageCenter, radius: radius));
    final imagePath = Path()
      ..addOval(
          Rect.fromCircle(center: imageCenter, radius: radius - borderWidth));
    final borderPath =
        Path.combine(PathOperation.difference, outlinePath, imagePath);

    final dotCenter = imageCenter.translate(0, radius + spacing + dotRadius);
    final dotPath = Path()
      ..addOval(
        Rect.fromCircle(center: dotCenter, radius: dotRadius),
      );

    final reducePath = Path.combine(PathOperation.union, outlinePath, dotPath);

    canvas
      ..clipPath(reducePath)
      ..drawImage(image, imageStartingPoint, Paint())
      ..drawPath(borderPath, Paint()..color = params.color)
      ..drawCircle(dotCenter, dotRadius, Paint()..color = params.color);

    final recorded = await recorder
        .endRecording()
        .toImage(diameter.toInt(), (diameter + spacing + dotDiameter).toInt());
    final byteData = await recorded.toByteData(format: ImageByteFormat.png);
    final result = byteData!.buffer.asUint8List();
    await cacheManager.putFile(
      url,
      result,
      key: cacheKey,
      fileExtension: 'png',
    );
    return result;
  }
}
